/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#include <time.h>
#include <string.h>

#if !defined(_WIN32)
#   include <strings.h>
#endif

#include "lcd.h"
#include "module.h"
#include "screen.h"

LCD_API int ModuleCreate(int handle,
                         void ** context,
                         LPMODULE_FN_ARRAY pFnArray);

LCD_API int ModuleDestroy(void * context);


LCD_API int GetDisplayData(void * context,
                           LPSCREEN pScreen,
                           unsigned int row,
                           char ** buf);


typedef struct _tag_screen_fans
{
    /** Number of fans being monitored */
    unsigned int nFanCount;

    /** Array of unsigned ints to contain rpm values */
    unsigned int * rpmList;

    /** Scratch space to keep our data stored */
    char ** pszOutput;

} SCREEN_FANS, * LPSCREEN_FANS;


/* ----- Internal function definitions ----- */


int fanSpeedReport(void * context, 
                   unsigned int fan_id, 
                   unsigned int fan_speed);


/* ----- Implementation ----- */



int ModuleCreate(int handle, void ** context, LPMODULE_FN_ARRAY pFnArray)
{
    int rc = 0;

    LPSCREEN_FANS pInst = NULL;

    /* Create our own scratch space */
    pInst = malloc(sizeof(SCREEN_FANS));
    if (pInst == NULL) {
        printf("malloc() failed in Fans module create");
        rc = -1;
        goto end_of_function;
    }

    /* Default to looking after four fans */
    pInst->nFanCount = 4;

    /* Zero out the array contents */
    pInst->rpmList = calloc(pInst->nFanCount, sizeof(unsigned int));

    /* --- Populate the function array data --- */

    /* We can use fn array version 1 */
    pFnArray->version = 1;

    /* Setup our various functions */
    pFnArray->pfnTextData = GetDisplayData;
    pFnArray->pfnFanSpeedReport = fanSpeedReport;

    /* --- */

    /* Return context data back to the caller */
    *context = pInst;

end_of_function:

    return rc;
}


int ModuleDestroy(void * context)
{
    LPSCREEN_FANS pInst = (LPSCREEN_FANS) context;

    free(pInst);

    return 0;
}


int GetDisplayData(void * context, LPSCREEN pScreen, unsigned int row, char ** buf)
{
    /* Return code */
    int rc = 0;

    /* This module instance */
    LPSCREEN_FANS pInstance = NULL;

    /* Screen dimensions */
    unsigned int width = 20, height = 4;
    
    /* Loop var */
    unsigned int i = 0;

    printf("GetDisplayData[fans]\n");

    /* Grab the context we are working in */
    pInstance = (LPSCREEN_FANS) context;

    /* How many lines are we allowed? */
    //screenGetSize(pScreen, &width, &height);

    /* Alloc some memory to use as scratch space */
    if (pInstance->pszOutput == NULL) {
        pInstance->pszOutput = malloc(sizeof(char *) * height);
        for (i = 0; i < height; i++) {
            pInstance->pszOutput[i] = malloc(sizeof(char) * width);
            memset(pInstance->pszOutput[i], (int) '*', width);
        }
    }

    /* Make sure we don't try to monitor more fans than we can display */
    if (pInstance->nFanCount > height) {
        pInstance->nFanCount = height;
    }

    /* Return a pointer to the buffer we require displayed on this row */
    *buf = pInstance->pszOutput[row];

    return rc;
}

int render(void * context)
{
    LPSCREEN_FANS pInst = NULL;
    int rc = 0;
    
    unsigned int i = 0;
        
    /* Grab the instance data */
    pInst = (LPSCREEN_FANS) context;

    /* Work through each of the fans we look after and write out data as text */
    for (i = 0; i < pInst->nFanCount; i++) {

        unsigned int lastReport = pInst->rpmList[i];
        char * pszOutBuf = pInst->pszOutput[i];

        /* Is is stopped? */
        if (lastReport == 0) {
            sprintf(pszOutBuf, "FAN %u: STOPPED", 0);
        }
        else if (lastReport == 1) {
            sprintf(pszOutBuf, "FAN %u: SLOW", 0);
        }
        else if (lastReport == 0xFFFFFFFF) {
            sprintf(pszOutBuf, "FAN %u: ---- RPM", 0);
        }
        else {
            /* Clean up the last report to a sensible value */
            sprintf(pszOutBuf, "FAN %u: %4u RPM", 0, lastReport);
        }
    }

    return rc;
}


int fanSpeedReport(void * context, unsigned int fan_id, unsigned int fan_speed)
{
    LPSCREEN_FANS pInst = NULL;
        
    /* Grab the instance data */
    pInst = (LPSCREEN_FANS) context;

    /* Save the RPM data in some shared storage (round to nearest 5) */
    fan_speed = (((fan_speed * 2) + 5) / 10) * 5;
    pInst->rpmList[fan_id] = fan_speed;

    printf("*********** FAN SPEED REPORT %u @ %u *****************\n", fan_id, fan_speed);
    
    return 0;
}

