/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#include <errno.h>
#include <fcntl.h>
#include <string.h>

#if defined(_WIN32)
#   include "pthread_w32.h"
#   include "time_w32.h"
#else
#   include <time.h>
#   include <unistd.h>
#   include <termios.h>
#   include <sys/select.h>
#   include <sys/time.h>
#   include <pthread.h>
#endif

#include "lcd.h"
#include "trace.h"
#include "device.h"
#include "packet.h"
#include "display.h"
#include "module.h"

#define IILC_MODULE MODULE_TYPE_APP


typedef int (* pfnPacketFlatten)(LPPACKET pPacket,
                                 unsigned char * buf,
                                 int * pDataLen);

typedef int (* pfnPacketInflate)(unsigned char * buf,
                                 int bufLen,
                                 int * pBytesRead,
                                 LPPACKET * ppPacket);

typedef int (* pfnPacketCreate)(PACKET_TYPE type,
                                LPPACKET * ppPacket);



typedef struct _tag_device
{
#if defined(_WIN32)
    HANDLE Device;
#else
    int Device;
#endif

    /* ID of the RxThread */
    pthread_t        RxThreadID;

    /* Mutex to lock when waiting for an Rx packet */
    pthread_mutex_t  RxMutex;

    /* Semaphore on which to wait when waiting for an Rx packet */
    pthread_cond_t   RxSemaphore;

    /* Flag to indicate the Rx thread should stay running */
    unsigned int     RxRunning;

    /* The packet received in reply */
    LPPACKET         pReplyPacket;

    struct
    {
        pfnPacketFlatten     Flatten;

        pfnPacketInflate     Inflate;

        pfnKeyEventListener  KeyEvent;

        pfnTempReportListener TempReport;

        pfnFanReportListener    FanSpeedReport;

    } pfn;

    struct
    {
        void * FanSpeedReport;

        void * KeyEvent;

        void * TempReport;
    
    } context;

    char *     pszPort;


} DEVICE;


LCD_API int deviceWritePacket(LPDEVICE pDevice,
                               LPPACKET pPacket,
                               LPPACKET * ppReplyPacket);



static void * RxThread(void * arg);

int deviceReceiveTempReport(LPDEVICE pDevice,
                             LPPACKET pPacket);

int deviceReceiveFanSpeedReport(LPDEVICE pDevice,
                                 LPPACKET pPacket);

int deviceReceiveKeyPressReport(LPDEVICE pDevice, 
                                 LPPACKET pPacket);


int deviceSetFanFailsafeTimeout(LPDEVICE pDevice, 
                                unsigned int mask, 
                                int milliseconds);

int deviceCreate(const char * pszClass, const char * pszPort, LPDEVICE * ppDevice)
{
    int rc = 0;

    LPDEVICE pDevice = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_deviceCreate);
    traceDebug(IILC_TRACE_CONTEXT,
        "pszClass %p, pszPort %p, ppDevice %p", pszClass, pszPort, ppDevice);

    if (pszClass == NULL || pszPort == NULL || ppDevice == NULL) {
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    traceDebug(IILC_TRACE_CONTEXT,
        "Creating device of type [%s] on port [%s]", pszClass, pszPort);

    /* Allocate the main memory */
    pDevice = (LPDEVICE) malloc(sizeof(DEVICE));

    /* Take a copy of the port string */
    pDevice->pszPort = (char *) malloc(strlen(pszPort)+1);
    strcpy(pDevice->pszPort, pszPort);

    pDevice->pfn.Flatten = packetFlatten;
    pDevice->pfn.Inflate = packetInflate;

    /* Mutex to lock when waiting for an Rx packet */
    pthread_mutex_init(&(pDevice->RxMutex), NULL);

    /* Semaphore on which to wait when waiting for an Rx packet */
    pthread_cond_init(&(pDevice->RxSemaphore), NULL);

    /* Flag to indicate the Rx thread should stay running */
    pDevice->RxRunning = 1;

    /* The packet received in reply */
    pDevice->pReplyPacket = NULL;

    /* TODO: Load the correct device driver module */

    /* Assign to user area */
    *ppDevice = pDevice;
    
end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}


int deviceDispose(LPDEVICE * ppDevice)
{
    int rc = 0;
    LPDEVICE pDev = NULL;

    if (ppDevice == NULL) {
        goto end_of_function;
    }

    pDev = *ppDevice;

    if (pDev->pszPort != NULL) {
        free(pDev->pszPort);
    }

    free(pDev);

end_of_function:

    return rc;
}


int deviceStart(LPDEVICE pDevice)
{
	int rc = 0;

#if defined(_WIN32)
    DCB          dcb;
    COMMTIMEOUTS our_timeouts;
#else
	struct termios portset;
#endif
	
    IILC_TRACE_ENTRY(IILC_MODULE, T_deviceStart);
    traceDebug(IILC_TRACE_CONTEXT, "pDevice %p", pDevice); 

    if (pDevice == NULL) {
        traceError(IILC_TRACE_CONTEXT, "pDevice == NULL");
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    traceDebug(IILC_TRACE_CONTEXT, "port %s", pDevice->pszPort);

#if defined(_WIN32)

    /* Open the USB serial port */
    pDevice->Device = CreateFile(pDevice->pszPort, GENERIC_READ | GENERIC_WRITE,
        0, NULL, OPEN_EXISTING, 0, NULL);

	if (pDevice->Device == INVALID_HANDLE_VALUE) {
        traceDebug(IILC_TRACE_CONTEXT, "Could not open port: %s", strerror(errno));
		rc = 1;
		goto end_of_function;
	}
    
    /*
     * Omit the call to SetupComm to use the default queue sizes.
     * Get the current configuration.
     */
    rc = GetCommState(pDevice->Device, &dcb);
    if (!rc) {
        traceDebug(IILC_TRACE_CONTEXT, "Could not get comm state");
        goto end_of_function;
    }

    /* Fill in the DCB: baud=115200, 8 data bits, no parity, 1 stop bit. */
    dcb.BaudRate = 115200;
    dcb.ByteSize = 8;
    dcb.Parity = NOPARITY;
    dcb.StopBits = ONESTOPBIT;
    dcb.fRtsControl = RTS_CONTROL_ENABLE;
    dcb.fDtrControl = DTR_CONTROL_ENABLE; 
    dcb.fAbortOnError = FALSE; 
    
    rc = SetCommState(pDevice->Device, &dcb);
    if (!rc) {
        traceDebug(IILC_TRACE_CONTEXT, "Could not set comm state");
        goto end_of_function;
    }

    if (GetCommTimeouts(pDevice->Device, &our_timeouts) != TRUE) {
        traceDebug(IILC_TRACE_CONTEXT, "Could not get comm timeouts");
        goto end_of_function;
    }

    /* 9600 is 1 byte per mS. Set it to 2 mS so the read will stay
     * as long as data is coming in. Will wait at least 100mS for
     * the 635 to respond.
     */
    our_timeouts.ReadIntervalTimeout        = 2000;
    our_timeouts.ReadTotalTimeoutMultiplier = 2000;  /* 1 second */
    our_timeouts.ReadTotalTimeoutConstant   = 2000;

    if (SetCommTimeouts(pDevice->Device, &our_timeouts) != TRUE) {
        traceDebug(IILC_TRACE_CONTEXT, "Could not set comm timeouts");
        goto end_of_function;
    }

#else

    /* Open the USB serial port */
    pDevice->Device = open(pDevice->pszPort, O_RDWR | O_NOCTTY | O_NDELAY);
	if (pDevice->Device == -1) {
        traceDebug(IILC_TRACE_CONTEXT, "Could not open port: %s", strerror(errno));
		rc = 1;
		goto end_of_function;
	}

	if (tcgetattr(pDevice->Device, &portset) == -1) {
		traceDebug(IILC_TRACE_CONTEXT, "tcgetattr() failed: %s", strerror(errno));
		rc = 1;
		goto end_of_function;
	}

	cfmakeraw(&portset);
	portset.c_cflag |= 0;
	cfsetispeed(&portset, B115200);
	cfsetospeed(&portset, B115200);
	if (tcsetattr(pDevice->Device, TCSANOW, &portset) == -1) {
		printf("tcsetattr() failed: %s\n", strerror(errno));
		rc = 1;
		goto end_of_function;
	}

#endif

    /* Pause required to allow device to initialise */
	SLEEP(350);

    /* Spin off a receiving thread */
    rc = pthread_create(&(pDevice->RxThreadID), NULL, RxThread, (void *) pDevice);
    if (rc) {
        printf("pthread_create error: %s\n", strerror(errno));
        goto end_of_function;
    }

    /* Any init that must happen first should go here before the Tx thread kicks off */
    deviceSetFanFailsafeTimeout(pDevice, 15, 5000);

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
	return rc;
}


int deviceStop(LPDEVICE pDevice)
{

    deviceSetFanFailsafeTimeout(pDevice, 0, 5000);

    close(pDevice->Device);

    return 0;
}





int deviceSetFanFailsafeTimeout(LPDEVICE pDevice, unsigned int mask, int milliseconds)
{
    int rc = 0, timeout = 0;
    PACKET_TYPE replyType = 0;
    LPPACKET pPacket = NULL, pReplyPacket = NULL;

    /* Calculate how many 1/8 second are in 'milliseconds' (1/8s == 125ms) */
    timeout = milliseconds / 125;

    /* Ensure we pass a good value in */
    if (timeout == 0)       { timeout++; }
    else if (timeout > 255) { timeout = 255; }

    rc = packetCreate(PACKET_TYPE_SET_FAN_FAILSAFE, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, mask);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, timeout);
    if (rc) { goto end_of_function; }

    rc = deviceWritePacket(pDevice, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

    rc = packetGetType(pReplyPacket, &replyType);
    if (replyType != (PACKET_TYPE_SET_FAN_FAILSAFE | 0x40)) {
        printf("Bad packet ack received for temperature reporting (%d)\n", replyType);
        rc = -1;
    }

end_of_function:

    /* Clean up resources */
    if (pPacket != NULL) {
        (void) packetDispose(&pPacket);
    }

    if (pReplyPacket != NULL) {
        (void) packetDispose(&pReplyPacket);
    }

    return rc;
}



int deviceWritePacket(LPDEVICE pDevice, LPPACKET pPacket, LPPACKET * ppReplyPacket)
{
    int rc = 0;
    int run, ret;

    struct timespec timeToWait;

    unsigned char buf[1024];
    int bufLen = sizeof(buf) / sizeof(buf[0]);
    int dataLen = bufLen;

    LPPACKET pReplyPacket = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_deviceWritePacket);
    traceDebug(IILC_TRACE_CONTEXT,
        "pDevice %p, pPacket %p, ppReplyPacket %p", pDevice, pPacket, ppReplyPacket);

    IILC_ASSERT(pDevice != NULL);
    IILC_ASSERT(pPacket != NULL);

    /* Flatten the packet into a buffer */
    traceDebug(IILC_TRACE_CONTEXT, "pDevice->pfnFlatten %p", pDevice->pfn.Flatten);

    rc = packetFlatten(pPacket, buf, &dataLen);
    if (rc) {
        traceDebug(IILC_TRACE_CONTEXT, "Failure while flattening packet %d", rc);
        goto end_of_function;
    }


    /* ----- Send it to the device ----- */

	/* Repeat this operation a few times */
	for (run = 0; run < 10; run++) {

		ret = write(pDevice->Device, buf, dataLen);
		if (ret < 0) {
			traceDebug(IILC_TRACE_CONTEXT, "write %d %s", ret, strerror(errno));
		}
		else {
			break;
		}
		if (run > 0) {
			traceDebug(IILC_TRACE_CONTEXT, "write(): EAGAIN #%d", run);
		}

		SLEEP(1);
	}
	
	if (ret < 0) {
		traceDebug(IILC_TRACE_CONTEXT, "write() failed; %s", strerror(errno));
        goto end_of_function;
	}
    else if (ret != dataLen) {
		traceDebug(IILC_TRACE_CONTEXT, "partial write(): len=%d ret=%d", dataLen, ret);
        goto end_of_function;
	}


    /* ----- Now wait for a response packet ----- */


    /* Lock the mutex */
    rc = pthread_mutex_lock(&pDevice->RxMutex);
    if (rc) {
        printf("Rx mutex lock failed: %s\n", strerror(errno));
        goto end_of_function;
    }

    /* Wait for 250ms (which is what the PDF user doc says) */
    {
        unsigned int milliseconds = 250;

        clock_gettime(CLOCK_REALTIME, &timeToWait);

        timeToWait.tv_nsec += (milliseconds % 1000) * 1000000;
        timeToWait.tv_sec  += (milliseconds / 1000);

        /* Check within range */
        if (timeToWait.tv_nsec > 1000000000) {
            timeToWait.tv_sec++;
            timeToWait.tv_nsec -= 1000000000;
        }
    }

    /* Wait on the semaphore */
    rc = pthread_cond_timedwait(&(pDevice->RxSemaphore), &(pDevice->RxMutex), &timeToWait);
    if (rc == ETIMEDOUT) {
        printf("Timed-out waiting for return packet\n");
        goto end_of_function;
    }
    else if (rc) {
        printf("Wait on semaphore failed: %s\n", strerror(errno));
        goto end_of_function;
    }

    /* Grab the returned packet and clear */
    pReplyPacket = pDevice->pReplyPacket;
    pDevice->pReplyPacket = NULL;

    /* Release the Rx mutex */
    rc = pthread_mutex_unlock(&(pDevice->RxMutex));
    if (rc) {
        printf("Rx mutex unlock failed: %s\n", strerror(errno));
        goto end_of_function;
    }


    /* ----- Return to the user ----- */


    /* Does the user wish to know about the returned packet? */
    if (ppReplyPacket != NULL) {
        *ppReplyPacket = pReplyPacket;
    }
    else {
        /* User would never know about the packet - dispose ourselves */
        if (pReplyPacket != NULL) {
            packetDispose(&pReplyPacket);
        }
    }
	
end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}


/*
 * Continuously-running thread which handles comms coming from the device.
 * Dispatches the received packets either back to the calling thread, or
 * to handlers of async events.
 */
void * RxThread(void * arg)
{
    LPDEVICE pDevice = (LPDEVICE) arg;

    int bytesRead = 0;
    int bytesAvail = 0;
    int ret;

    unsigned char buf[1024];
    int bufLen = sizeof(buf) / sizeof(buf[0]);

	do {
		int rc = 0;
		LPPACKET pReplyPacket = NULL;

		/* Read from the port */
		ret = read(pDevice->Device, &buf[bytesAvail], bufLen - bytesAvail);
        if (ret < 0 && errno == EAGAIN) {
            /* Snooze for 10ms while we wait for something to turn up */
            SLEEP(10);
            continue;
        }

        /* We now have more in the buffer */
        bytesAvail += ret;

        /* Nothing has been read from the buffer yet */
        bytesRead = 0;

        do {

            /* Attempt to create a packet from bytes avaialble in this buffer */
            rc = (*pDevice->pfn.Inflate)(buf, bytesAvail, &bytesRead, &pReplyPacket);

            /* Remove any bytes that have been read */
            if (bytesRead > 0) {
                bytesAvail -= bytesRead;
                memmove(buf, buf + bytesRead, bytesAvail);
            }

            /* Have we succeeded? */
            if (rc == 0) {

                /* What sort of packet is this? */
                int packetClass = packetGetClass(pReplyPacket);
                switch (packetClass) {

                /* Async report from device to app */
                case PACKET_CLASS_NORMAL_REPORT:
                {
                    /* Get the packet type */
                    PACKET_TYPE packetType;
                    packetGetType(pReplyPacket, &packetType);

                    /* Decide what to do with this */
                    switch (packetType) {

                    case PACKET_TYPE_RPT_TEMP_SENSOR:
                        deviceReceiveTempReport(pDevice, pReplyPacket);
                        break;

                    case PACKET_TYPE_RPT_FAN_SPEED:
                        deviceReceiveFanSpeedReport(pDevice, pReplyPacket);
                        break;

                    case PACKET_TYPE_RPT_KEY_ACTIVITY:
                        deviceReceiveKeyPressReport(pDevice, pReplyPacket);
                        break;

                    default:
                        printf("Unknown packet type (%X)\n", packetType);
                        break;
                    }

                    /* Dispose of the packet */
                    packetDispose(&pReplyPacket);
                    break;
                }

                /* Both normal and error responses to be handed back to waiting caller */
                case PACKET_CLASS_NORMAL_RESPONSE:
                case PACKET_CLASS_ERROR_RESPONSE:
                {
                    PACKET_TYPE ptype;
                    packetGetType(pReplyPacket, &ptype);

                    /* Lock the Rx mutex */
                    rc = pthread_mutex_lock(&pDevice->RxMutex);
                    if (rc) {
                        printf("Rx mutex lock failed: %s\n", strerror(errno));
                        break;
                    }

                    /* Set the reply packet (memory now owned by the device) */
                    pDevice->pReplyPacket = pReplyPacket;

                    /* Signal the Rx semaphore */
                    rc = pthread_cond_signal(&(pDevice->RxSemaphore));
                    if (rc) {
                        printf("Rx semaphore signal failed: %s\n", strerror(errno));
                        break;
                    }

                    /* Release the Rx mutex */
                    rc = pthread_mutex_unlock(&(pDevice->RxMutex));
                    if (rc) {
                        printf("Rx mutex unlock failed: %s\n", strerror(errno));
                        break;
                    }
                    break;
                }

                /* Shouldn't get this packet received! */
                case PACKET_CLASS_NORMAL_REQUEST:
                    printf("Strange packet type received\n");

                    /* Dispose of the packet */
                    packetDispose(&pReplyPacket);
                    break;

                /* Broken!! */
                default:
                    printf("Unable to determine packet class\n");

                    /* Dispose of the packet */
                    packetDispose(&pReplyPacket);
                    break;
                }
            }
            else if (rc == ERR_INSUFFICIENT_DATA) {
                /* Just a question of waiting for more? */
                SLEEP(10);
                break; /* out of bytesAvail loop */
            }
            else {
                /* Must be bad */
                printf("Fatal error from packet inflate %d\n", rc);
                break;
            }

        } while (bytesAvail > 0);

    } while (pDevice->RxRunning);
	
	return NULL;
}

int deviceSetBacklight(LPDEVICE pdevice, unsigned char backlightValue)
{
    int rc = 0;
    LPPACKET pPacket = NULL, pReplyPacket = NULL;

    rc = packetCreate(PACKET_TYPE_SET_BACKLIGHT, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, backlightValue);
    if (rc) { goto end_of_function; }

    rc = deviceWritePacket(pdevice, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

end_of_function:

    if (pPacket != NULL) {
        (void) packetDispose(&pPacket);
    }
    if (pReplyPacket != NULL) {
        (void) packetDispose(&pReplyPacket);
    }

    return rc;
}

int deviceSetContrast(LPDEVICE pdevice, unsigned char contrastValue)
{
    int      rc      = 0;
    LPPACKET pPacket = NULL;

    rc = packetCreate(PACKET_TYPE_SET_LCD_CONTRAST, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, contrastValue);
    if (rc) { goto end_of_function; }

    rc = deviceWritePacket(pdevice, pPacket, NULL);
    if (rc) { goto end_of_function; }

end_of_function:

    if (pPacket != NULL) { packetDispose(&pPacket); }

    return rc;
}


int deviceClear(LPDEVICE pdevice)
{
    int rc = 0;
    LPPACKET pPacket = NULL;
    LPPACKET pReplyPacket = NULL;

    rc = packetCreate(PACKET_TYPE_CLEAR_SCREEN, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = deviceWritePacket(pdevice, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

    rc = packetDispose(&pPacket);
    if (rc) { goto end_of_function; }

end_of_function:

    return rc;
}



int deviceWriteText(LPDEVICE pdevice, int x, int y, char * strData, int width)
{
    int rc = 0;
    LPPACKET pPacket = NULL, pReplyPacket = NULL;

    rc = packetCreate(31, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, x);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, y);
    if (rc) { goto end_of_function; }

    rc = packetWriteString(pPacket, strData, width);
    if (rc) { goto end_of_function; }

    rc = deviceWritePacket(pdevice, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

end_of_function:

    if (pPacket) { packetDispose(&pPacket); }
    if (pReplyPacket) { packetDispose(&pReplyPacket); }

    return rc;
}



int deviceReboot(LPDEVICE pdevice)
{
    int rc = 0;
    LPPACKET pPacket = NULL;

    rc = packetCreate(PACKET_TYPE_POWER_CONTROL, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, 8);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, 18);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, 99);
    if (rc) { goto end_of_function; }

    rc = deviceWritePacket(pdevice, pPacket, NULL);
    if (rc) { goto end_of_function; }

end_of_function:

    /* Clean up resources */
    if (pPacket != NULL) {
        (void) packetDispose(&pPacket);
    }

    return rc;
}

int deviceSetFanPower(LPDEVICE pdevice, unsigned char fan1, unsigned char fan2,
                       unsigned char fan3, unsigned char fan4)
{
    int rc = 0;
//    PACKET_TYPE replyType = 0;
    LPPACKET pPacket = NULL, pReplyPacket = NULL;

    rc = packetCreate(PACKET_TYPE_SET_FAN_POWER, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, fan1);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, fan2);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, fan3);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, fan4);
    if (rc) { goto end_of_function; }

    rc = deviceWritePacket(pdevice, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

    /*
    rc = packetGetType(pReplyPacket, &replyType);
    if (replyType != (PACKET_TYPE_SET_TEMP_REPORTING | 0x40)) {
        printf("Bad packet ack received for temperature reporting (%d)\n", replyType);
        rc = -1;
    }
    */

end_of_function:

    /* Clean up resources */
    if (pPacket != NULL) {
        (void) packetDispose(&pPacket);
    }

    if (pReplyPacket != NULL) {
        (void) packetDispose(&pReplyPacket);
    }

    return rc;
}


int deviceSetFanReporting(LPDEVICE pdevice, unsigned int mask)
{
    int rc = 0;
    //PACKET_TYPE replyType = 0;
    LPPACKET pPacket = NULL, pReplyPacket = NULL;

    rc = packetCreate(PACKET_TYPE_SET_FAN_REPORTING, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, mask);
    if (rc) { goto end_of_function; }

    rc = deviceWritePacket(pdevice, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

    /*
    rc = packetGetType(pReplyPacket, &replyType);
    if (replyType != (PACKET_TYPE_SET_TEMP_REPORTING | 0x40)) {
        printf("Bad packet ack received for temperature reporting (%d)\n", replyType);
        rc = -1;
    }
    */

end_of_function:

    /* Clean up resources */
    if (pPacket != NULL) {
        (void) packetDispose(&pPacket);
    }

    if (pReplyPacket != NULL) {
        (void) packetDispose(&pReplyPacket);
    }

    return rc;
}

int deviceSetTemperatureReporting(LPDEVICE pdevice, unsigned int mask)
{
    int rc = 0;
    //PACKET_TYPE replyType = 0;
    LPPACKET pPacket = NULL, pReplyPacket = NULL;

    rc = packetCreate(PACKET_TYPE_SET_TEMP_REPORTING, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, mask & 0x000000FF);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, (mask & 0x0000FF00) >> 8);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, (mask & 0x00FF0000) >> 16);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, (mask & 0xFF000000) >> 24);
    if (rc) { goto end_of_function; }

    rc = deviceWritePacket(pdevice, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

    /*
    rc = packetGetType(pReplyPacket, &replyType);
    if (replyType != (PACKET_TYPE_SET_TEMP_REPORTING | 0x40)) {
        printf("Bad packet ack received for temperature reporting (%d)\n", replyType);
        rc = -1;
    }
    */

end_of_function:

    /* Clean up resources */
    if (pPacket != NULL) {
        (void) packetDispose(&pPacket);
    }

    if (pReplyPacket != NULL) {
        (void) packetDispose(&pReplyPacket);
    }

    return rc;
}

/**
 * Called when a fan speed report is received from the connected display.
 */
int deviceReceiveFanSpeedReport(LPDEVICE pDevice, LPPACKET pPacket)
{
    int rc = 0;

    /* Final fan speed value */
    unsigned int fanSpeed = 0;
    
    /* Raw data from the packet */
    unsigned char LSB, MSB, fan_number, number_of_fan_tach_cycles;

    /* Pulses Per Revolution  - specific to each fan, most commonly 2 */
    const int PPR = 2;

    /* Join of MSB+LSB */
    unsigned short Fan_Timer_Ticks;

    IILC_TRACE_ENTRY(IILC_MODULE, T_displayReceiveFanSpeedReport);
    traceDebug(IILC_TRACE_CONTEXT,
        "pDevice " T_FMT_PTR ", pPacket " T_FMT_PTR,
        pDevice, pPacket);

    /* Do we have any listeners? */
    if (pDevice->pfn.FanSpeedReport == NULL) {
        goto end_of_function;
    }

    /* Read the raw data from the packet */
    if (!rc) { rc = packetReadData(pPacket, &fan_number); }
    if (!rc) { rc = packetReadData(pPacket, &number_of_fan_tach_cycles); }
    if (!rc) { rc = packetReadData(pPacket, &LSB); }
    if (!rc) { rc = packetReadData(pPacket, &MSB); }
    if (rc) {
        traceWarning(IILC_TRACE_CONTEXT, "Badly-formed fan speed report packet");
        goto end_of_function;
    }

    /* Convert raw data to something useful */
    Fan_Timer_Ticks = (MSB << 8) + LSB;

    /* Useful information dumped our to trace */
    traceDebug(IILC_TRACE_CONTEXT,
        "fan_number " T_FMT_UINT " Fan_Timer_Ticks " T_FMT_UINT " Tach_Cycles " T_FMT_UINT,
        (unsigned int) fan_number, (unsigned int) Fan_Timer_Ticks, number_of_fan_tach_cycles);

    if (number_of_fan_tach_cycles <= 3) {        
        fanSpeed = 0; /* Stopped */
    }
    else if (number_of_fan_tach_cycles <= 4) {        
        fanSpeed = 1; /* Too slow to measure */
    }
    else if (number_of_fan_tach_cycles == 0xFF) {
        fanSpeed = 0XFFFFFFFF; /* Out of range */
    }
    else {
        /* Use the formula provided by CF-635 data sheet v 1.4 */
        fanSpeed = ((27692308L / PPR) * (unsigned long)(number_of_fan_tach_cycles - 3)) / (Fan_Timer_Ticks);
    }

    traceDebug(IILC_TRACE_CONTEXT, "fan speed " T_FMT_UINT, fanSpeed);

    /* Send this data to the callback */
    (pDevice->pfn.FanSpeedReport)(pDevice->context.FanSpeedReport, fan_number, fanSpeed);

end_of_function:

    IILC_TRACE_EXIT;
    return rc;
}


int deviceReceiveKeyPressReport(LPDEVICE pDevice, LPPACKET pPacket)
{
    unsigned char event;

    IILC_TRACE_ENTRY(IILC_MODULE, T_displayReceiveKeyPressReport);
    traceDebug(IILC_TRACE_CONTEXT,
        "pDevice " T_FMT_PTR ", pPacket " T_FMT_PTR, pDevice, pPacket);

    /* Ignore if no-one is bothered about receiving reports */
    if (pDevice == NULL) {
        goto end_of_function;
    }

    /* Read the data */
    packetReadData(pPacket, &event);

    /* Perform the callback */
    (pDevice->pfn.KeyEvent)(pDevice->context.KeyEvent, (PACKET_KEY_EVENT) event);

end_of_function:

    IILC_TRACE_EXIT;
    return 0;
}


/**
 * Called when a temperature report is received from the connected display.
 */
int deviceReceiveTempReport(LPDEVICE pDevice, LPPACKET pPacket)
{
    int rc = 0;
    //int i = 0;
    double degc = 0.0;
    unsigned char sensor, MSB, LSB, DOW_crc_status;
    unsigned short report;

    /* Do we have any callbacks registered? */
    if (pDevice->pfn.TempReport == NULL) {
        goto end_of_function;
    }

    /* Read the raw data */
    packetReadData(pPacket, &sensor);
    packetReadData(pPacket, &LSB);
    packetReadData(pPacket, &MSB);
    packetReadData(pPacket, &DOW_crc_status);

    /* Check CRC is OK */
    if (DOW_crc_status == 0) {
        printf("Bad DOW CRC\n");
        rc = -1;
        goto end_of_function;
    }

    /* Convert to a full 16-bit value and get true reading */
    report = (MSB << 8) + LSB;
    degc = ((double) report) / 16.0; 

    (pDevice->pfn.TempReport)(pDevice->context.TempReport, sensor, degc);

end_of_function:

    return rc;
}



int deviceSetLED(LPDEVICE pDevice, unsigned int led_id, LPCOLOUR pColour)
{
    int rc = 0;
    LPPACKET pPacket = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_displaySetLED);

    traceDebug(IILC_TRACE_CONTEXT,
        "pDevice " T_FMT_PTR ", led_id " T_FMT_UINT ", pColour " T_FMT_PTR,
        pDevice, led_id, pColour);

    /* Ignore the blue content of the colour */

    /* Write out the green component */
    packetCreate(34, &pPacket);
    packetWriteData(pPacket, 11 - (led_id * 2));
    packetWriteData(pPacket, pColour->green);
    deviceWritePacket(pDevice, pPacket, NULL);
    packetDispose(&pPacket);

    /* Write out the red component */
    packetCreate(34, &pPacket);
    packetWriteData(pPacket, 12 - (led_id * 2));
    packetWriteData(pPacket, pColour->red);
    deviceWritePacket(pDevice, pPacket, NULL);
    packetDispose(&pPacket);

//end_of_function:

    traceDebug(IILC_TRACE_CONTEXT, "rc " T_FMT_INT, rc);
    IILC_TRACE_EXIT;
    return (rc);
}

int deviceSetCallbackTemperature(LPDEVICE pDevice, pfnTempReportListener pfn, void * context)
{
    pDevice->pfn.TempReport = pfn;
    pDevice->context.TempReport = context;
    
    return 0;   
}

int deviceSetCallbackFan(LPDEVICE pDevice, pfnFanReportListener pfn, void * context)
{
    pDevice->pfn.FanSpeedReport = pfn;
    pDevice->context.FanSpeedReport = context;

    return 0;
}

int deviceSetCallbackKeyEvent(LPDEVICE pDevice, pfnKeyEventListener pfn, void * context)
{

    pDevice->pfn.KeyEvent = pfn;
    pDevice->context.KeyEvent = context;

    return 0;
}


