/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#include <errno.h>
#include <fcntl.h>
#include <string.h>

#if !defined(_WIN32)
#   include <time.h>
#   include <unistd.h>
#   include <termios.h>
#   include <sys/select.h>
#   include <sys/time.h>
#   include <pthread.h>
#endif

#include "lcd.h"
#include "trace.h"
#include "display.h"
#include "packet.h"
#include "eventQueue.h"
#include "device.h"
#include "module.h"
#include "vector.h"

/*
 * This file is the "display" module.
 */
#define IILC_MODULE MODULE_TYPE_DISPLAY


/* Constant properties of this display */
#define PHYSICAL_DEVICE_WIDTH      20
#define PHYSICAL_DEVICE_HEIGHT      4
#define NUMBER_OF_LEDS              4
#define NUMBER_OF_FANS              4



int dispatchReportPacket(LPDISPLAY pDisplay, LPPACKET pPacket);

void * TxThread(void * arg);

void * RenderThread(void * arg);


int displayRefreshScreen(LPDISPLAY pDisplay);



void displayCallbackTemp(void * context, unsigned int sensor, double degc);

void displayCallbackKeyEvent(void * context, PACKET_KEY_EVENT event);

void displayCallbackFan(void * context, unsigned int fan, unsigned int rpm);

typedef struct _tag_led_instance
{
    /** Colour of the LED when it is "on" */
    COLOUR on;

    /** Colour of the LED when it is "off" */
    COLOUR off;

    /** The percentage of time which the LED spends on */
    unsigned int dutyCycle;

    /** The frequency of cycling from off->on->off in Hz/100 */
    unsigned int freq;

    /** Internal flag to track if the LED is on or off */
    int bLEDOn;

    /** The callback function used to decide this LED's status */
    pfnGetLEDStatus pfn;

    /** The context to supply to the above callback function */
    void * context;

    /** The ID presented to the module controlling this LED */
    int led_id;

} LED_INSTANCE, * LPLED_INSTANCE;


typedef struct _tag_display_temp_listener
{
    /** Function to receive temperature report */
    pfnTempReportListener pfn;

    /** Context data to supply to the function */
    void * context;

    /** Sensor ID supplied to the callback */
    int sensor_id;

} TEMP_LISTENER, * LPTEMP_LISTENER;

typedef struct _tag_display_fan_listener
{
    /** Fan ID which this relates to */
    unsigned int fan_id;

    /** Function to receive fan speed report */
    pfnFanReportListener pfn;

    /** Context data to supply to the function */
    void * context;

    /** Fan ID which this is reported as */
    unsigned int reportAs;
}
FAN_LISTENER, * LPFAN_LISTENER;

typedef struct _tag_display
{
    /* ----- Physical characteristics of the display ----- */

    /* Physical width of the device */
    unsigned int width;

    /* Physical height of the device */
    unsigned int height;

    /* Number of LEDs physically available on this display */
    unsigned int nLED;

    /* Number of fans physically available on this display */
    unsigned int nFans;

    /** Number of temperature sensors connected */
    unsigned int nTempSensors;

    /* ----- LEDs ----- */

    /** List of data regions relating to each LED */
    LPLED_INSTANCE * pLEDInstanceList;

    /* ----- Fans ----- */

    LPVECTOR pFanListeners;

    /** Mask of fans to set reporting for */
    unsigned int fanMask;

    /* ----- Temperature sensors ----- */

    /** Number of listeners for each fan available */
    unsigned int * nTempListeners;

    /** List of temperature report listener structs for each temperature sensor */
    TEMP_LISTENER ** pTempListeners;

    /* ----- LEDs ----- */

    pfnKeyEventListener pfnKeyListener;

    void * pKeyListenerCtxt;


    /* ----- Screen information ----- */

    /** Number of screens to be displayed in the cycle */
    int nScreenCount;

    /** The current screen being displayed */
    int nScreenCurrent;

    /** The list of screens to be displayed */
    LPSCREEN * pScreenList;

    /** List of threads which control the rendering of each screen */
    pthread_t * screenThreads;

    /* ----- Others ----- */

    /* The event queue around which the whole Tx loop rotates */
    LPEVENT_QUEUE    pEventQueue;

    /* ID of the Txthread */
    pthread_t        TxThreadID;

    LPDEVICE pDevice;

} DISPLAY;


/**
 * Creates an instance of the display.
 */
int displayCreate(LPDEVICE pDevice, LPDISPLAY * ppDisplay)
{
    LPDISPLAY pDisplay = NULL;
    int rc = 0;
    unsigned int i = 0;

    /* Sanity check */
    if (ppDisplay == NULL) {
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

	/* Allocate some memory */
	pDisplay = (LPDISPLAY) malloc(sizeof(DISPLAY));
	if (pDisplay == NULL) {
		rc = ERR_MEM_ALLOC_FAILED;
		goto end_of_function;
	}

    /* Use the supplied device for I/O */
    pDisplay->pDevice = pDevice;

    /* Let the device know we want to receive all possible callbacks */
    deviceSetCallbackTemperature(pDevice, displayCallbackTemp, pDisplay);
    deviceSetCallbackFan(pDevice, displayCallbackFan, pDisplay);
    deviceSetCallbackKeyEvent(pDevice, displayCallbackKeyEvent, pDisplay);

    /* Basic display information */
    pDisplay->width  = PHYSICAL_DEVICE_WIDTH;
    pDisplay->height = PHYSICAL_DEVICE_HEIGHT;
    pDisplay->nLED   = NUMBER_OF_LEDS;
    pDisplay->nFans  = NUMBER_OF_FANS;
    
    /* LED output information */
    pDisplay->pLEDInstanceList = malloc(sizeof(LPLED_INSTANCE) * pDisplay->nLED);
    for (i = 0; i < pDisplay->nLED; i++) {
        pDisplay->pLEDInstanceList[i] = (LPLED_INSTANCE) calloc(1, sizeof(LED_INSTANCE));
    }

    /* Fan output information */
    vectorCreate(&(pDisplay->pFanListeners));

    /* Temperate listener information */
    pDisplay->nTempSensors = 32; /* Total number of sensors that can be connected */
    pDisplay->nTempListeners = malloc(sizeof(unsigned int) * pDisplay->nTempSensors);
    pDisplay->pTempListeners = malloc(sizeof(LPTEMP_LISTENER) * pDisplay->nTempSensors);
    for (i = 0; i < pDisplay->nTempSensors; i++) {
        pDisplay->nTempListeners[i] = 0;
        pDisplay->pTempListeners[i] = NULL;
    }

    /* Screen information */
    pDisplay->nScreenCount = 0;
    pDisplay->nScreenCurrent = 0;
    pDisplay->pScreenList = NULL;
    pDisplay->screenThreads = NULL;

    /* Key listener information */
    /* pfnKeyEventListener pfnKeyListener; */
    /* void * pKeyListenerCtxt; */

    /* Tx thread is initialised during Start() */

    /* The event queue around which the whole Tx loop rotates */
    eventQueueCreate(&(pDisplay->pEventQueue));

end_of_function:

	/* Assign to the user */
    if (!rc) {
	    *ppDisplay = pDisplay;
    }

    return rc;
}



int displayStart(LPDISPLAY pDisplay)
{
    int rc = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_displayStart);
    traceDebug(IILC_TRACE_CONTEXT, "pDisplay %p", pDisplay);

    /* Start the underlying device */
    deviceStart(pDisplay->pDevice);

    /* Spin off a transmit thread */
    rc = pthread_create(&(pDisplay->TxThreadID), NULL, TxThread, (void *) pDisplay);
    if (rc) {
        printf("pthread_create error: %s\n", strerror(errno));
        goto end_of_function;
    }

    {
        EVENT evt;
        EVENT_INIT(evt, EVENT_POLL_MODULE, NULL, 0);
        eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

        EVENT_INIT(evt, EVENT_SET_BACKLIGHT, NULL, 30);
        eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

        /* Start fan at 100% */
        EVENT_INIT(evt, EVENT_SET_FAN_STATE, NULL, 0);
        eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

        /* Update the screen regularly */
        EVENT_INIT(evt, EVENT_SCREEN_REFRESH, NULL, 0);
        eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);
    }


end_of_function:

    IILC_TRACE_EXIT_RC(rc);
	return rc;
}


/**
  * Informs the display of the function to use when checking the fan power
  * level.
  *
  * \param pDisplay Display instance to which this relates.
  * \param fan_id The fan number as found in the display which is to be controlled.
  * \param pfn Pointer to a function which will receive the fan power request.
  * Using NULL here stops the fan being controlled by any external logic.
  * \param context Context data to provide to the callback function. May be NULL.
  *
  * \retval Zero on success. Any other value indicates error.
  */
int displaySetFanController(LPDISPLAY pDisplay, int fan_id, pfnGetFanPower pfn, void * context)
{


    return 0;
}


/**
  * Informs the display of the function to use when checking the fan power
  * level.
  *
  * \param pDisplay Display instance to which this relates.
  * \param led_id The LED number as found in the display which is to be controlled.
  * \param pfn Pointer to a function which will receive the LED display status.
  * Using NULL here stops the LED being controlled by any external logic, and will
  * switch the LED off.
  * \param context Context data to provide to the callback function. May be NULL.
  *
  * \retval Zero on success. Any other value indicates error.
  */
int displaySetLEDController(LPDISPLAY pDisplay, int led_id, pfnGetLEDStatus pfn, void * context, int report_as_led)
{
    int rc = 0;
    EVENT evt;
    LPLED_INSTANCE pInst = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_displaySetLEDController);
    traceDebug(IILC_TRACE_CONTEXT,
        "pDisplay " T_FMT_PTR ", led_id " T_FMT_INT ", pfn " T_FMT_PTR
        ", context " T_FMT_PTR ", report_as_led " T_FMT_INT,
        pDisplay, led_id, pfn, context, report_as_led);

    /* Sanity checks */
    if (pDisplay == NULL) {
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    if (led_id > pDisplay->nLED) {
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    /* Convenience */
    pInst = pDisplay->pLEDInstanceList[led_id];

    /* Clear and init memory */
    pInst->pfn     = pfn;
    pInst->context = context;
    pInst->led_id  = report_as_led;

    /* Are we setting up or removing a controller? */
    if (pfn != NULL) {

        /* Remember to query this LED */
        EVENT_INIT(evt, EVENT_POLL_LED_STATUS, NULL, led_id);
        eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

        /* Remember to update this LED */
        EVENT_INIT(evt, EVENT_LED_UPDATE, NULL, led_id);
        eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);
    }

end_of_function:

    IILC_TRACE_EXIT;
    return 0;
}

int displayDestroy(LPDISPLAY * ppDisplay)
{

    return 0;
}

int displayStop(LPDISPLAY pDisplay)
{
	int rc = 0;
    //int ret;
    EVENT evt;
    LPEVENT_QUEUE pEventQueue = pDisplay->pEventQueue;

        /* ----- Restore display to sensible state ----- */

        /* Switch fans back on */
        EVENT_INIT(evt, EVENT_SET_FAN_STATE, NULL, 100);
        eventQueueEnqueue(pEventQueue, &evt, 0);

        /* Stop temperature reporting */
        EVENT_INIT(evt, EVENT_SET_TEMPERATURE_REPORTING, NULL, 0);
        eventQueueEnqueue(pEventQueue, &evt, 0);

        /* Stop fan reporting */
        EVENT_INIT(evt, EVENT_SET_FAN_REPORTING, NULL, 0);
        eventQueueEnqueue(pEventQueue, &evt, 0);

        /* Backlight off */
        EVENT_INIT(evt, EVENT_SET_BACKLIGHT, NULL, 0);
        eventQueueEnqueue(pEventQueue, &evt, 0);

    /* Terminate the Tx thread */
    EVENT_INIT(evt, EVENT_SHUTDOWN, NULL, 0);
    eventQueueEnqueue(pEventQueue, &evt, 0);

    /* Wait for the Tx thread to complete */
    pthread_join(pDisplay->TxThreadID, NULL);

    /* Shutdown the device itself */
    deviceStop(pDisplay->pDevice);

    /* Release this memory */
    free(pDisplay);

	/* Clear user's reference */
	pDisplay = NULL;

//end_of_function:

	return rc;
}

typedef enum _tag_window_type
{
    WINDOW_TYPE_TEXT,
    WINDOW_TYPE_SCROLL_TEXT,
    WINDOW_TYPE_WRAP_TEXT,
    WINDOW_TYPE_BAR,
    WINDOW_TYPE_LARGE_TEXT

} WINDOW_TYPE;

typedef struct _tag_window
{
    unsigned int x;

    unsigned int y;

    unsigned int width;

    unsigned int height;

    /* Type of window to display */
    WINDOW_TYPE type;

    /* Function used to obtain the data for the display */
    union _tag_window_pfn {
        pfnGetTextData    GetText;
        pfnGetNumericData GetNumeric;
    } pfn;

    /* Context data to supply to the callback */
    void * context;

} WINDOW;


int displayAddWindow(LPDISPLAY pDisplay, WINDOW window)
{
    int rc = 0;

    return rc;
}


/**
 * Registers a callback which will receive regular reports regarding
 * the speed of fans connected to the display.
 *
 * \param pDisplay Pointer to the display which is being modified.
 * \param fan_id The physical fan number connected to the display.
 * \param pfnListener Callback function to receive the fan reports.
 * \param context Context data to be supplied to the callback function.
 * \param report_as_id A value to supply to the callback as a way of
 *        abstracting the physical fan ID from the fan ID received by
 *        the callback.
 *
 * \retval Zero on success. All other values indicate error.
 */
int displayAddFanReportListener(LPDISPLAY pDisplay, unsigned int fanId,
                                pfnFanReportListener pfnListener, void * context, 
                                unsigned int report_as_id)
{
    int rc = 0;
    EVENT evt;
    LPFAN_LISTENER pListener = NULL;

    if (pfnListener == NULL) {
        rc = -1;
        goto end_of_function;
    }

    /* Set the new data */
    pListener = (LPFAN_LISTENER) malloc(sizeof(FAN_LISTENER));
    pListener->pfn      = pfnListener;
    pListener->context  = context;
    pListener->fan_id   = fanId;
    pListener->reportAs = report_as_id;
    vectorAdd(pDisplay->pFanListeners, pListener);

    /* Setting listener */
    pDisplay->fanMask = pDisplay->fanMask | (1 << fanId);

    /* Add an update for this fan to the event queue */
    EVENT_INIT(evt, EVENT_SET_FAN_REPORTING, NULL, pDisplay->fanMask);
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

end_of_function:

    return rc;
}

int displayRemoveFanReportListener(LPDISPLAY pDisplay, pfnFanReportListener pfnListener)
{
    return 0;
}

int displayAddTempReportListener(LPDISPLAY pDisplay, int sensor_id,
                                 pfnTempReportListener pfnListener,
                                 void * context, int report_as_id)
{
    int rc = 0, i = 0, nCount = 0, sensorMask = 0;
    TEMP_LISTENER * pNewList = NULL;
    EVENT evt;


    if (pfnListener == NULL) {
        rc = -1;
        goto end_of_function;
    }

    /* Get the original number of listeners */
    nCount = pDisplay->nTempListeners[sensor_id];

    /* Alloc a new block of storage */
    pNewList = malloc(sizeof(TEMP_LISTENER) * (nCount + 1));

    /* Copy all old data across, if any */
    if (pDisplay->pTempListeners[sensor_id] != NULL) {
        memcpy(pNewList, pDisplay->pTempListeners[sensor_id], sizeof(TEMP_LISTENER) * nCount);
    }

    /* Set the new data */
    pNewList[nCount].pfn       = pfnListener;
    pNewList[nCount].context   = context;
    pNewList[nCount].sensor_id = report_as_id;

    /* Increase the number of listeners */
    pDisplay->nTempListeners[sensor_id]++;

    /* Store */
    pDisplay->pTempListeners[sensor_id] = pNewList;

    /* Work out which sensors are required for reporting */
    for (i = pDisplay->nTempSensors - 1; i > 0; i--) {
        sensorMask += pDisplay->nTempListeners[i] > 0 ? 1 : 0;
        sensorMask <<= 1;
    }
    sensorMask += pDisplay->nTempListeners[i] > 0 ? 1 : 0;

    /* Add an update for this fan to the event queue */
    EVENT_INIT(evt, EVENT_SET_TEMPERATURE_REPORTING, NULL, sensorMask);
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

end_of_function:

    return rc;
}

int displayRemoveTempReportListener(LPDISPLAY pDisplay, pfnTempReportListener pfnListener)
{
    return 0;
}

int displayAddKeyEventListener(LPDISPLAY pDisplay, pfnKeyEventListener pfnListener, void * context)
{

    return 0;
}



/****************************************************************************
 *                                                                          *
 * This set of functions handles the case where packets are received from   *
 * the display asynchronously and dispatches them to any registered         *
 * for that report type.                                                    *
 *                                                                          *
 ****************************************************************************/



int dispatchReportPacket(LPDISPLAY pDisplay, LPPACKET pPacket)
{
//    PACKET_TYPE packetType = -1;

    /*
    IILC_TRACE_ENTRY((IILC_MODULE, T_dispatchReportPacket, "pPacket %p", pPacket));
    */


    return 0;
}

void * TxThread(void * arg)
{
    /* Grab our instance data from the arg ptr */
    LPDISPLAY pDisplay = (LPDISPLAY) arg;
    
    /* Vars used in this function */
    int running = 1;
    int rc = 0;

    /* Initialise convenient locals */
    LPEVENT_QUEUE pEventQueue = pDisplay->pEventQueue;

    /* Temp */
    char * longText = "Katrina you are so beautiful \x93\x93\x93";
    char scrData[1024];
    int scrollPos = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_TxThread);

    while (running)
    {
        /* Loop vars */
        EVENT evt;

        /* Flag to indicate we should keep the memory for the event (saves reallocing) */
        int iRequeueTime = -1; /* -1 == dispose of event */

        /* Wait for event dequeue */
        rc = eventQueueDequeue(pEventQueue, &evt);
        if (rc) {
            printf("Event dequeue failed\n");
            break;
        }

        /* Process */
        switch (evt.eType)
        {
        case EVENT_SHUTDOWN:
            running = 0;
            break;

        case EVENT_SCREEN_CLEAR:
            deviceClear(pDisplay->pDevice);
            break;

        case EVENT_SET_BACKLIGHT:
            deviceSetBacklight(pDisplay->pDevice, (unsigned char) evt.iData);
            break;

        case EVENT_SET_CONTRAST:
            deviceSetContrast(pDisplay->pDevice, (unsigned char) evt.iData);
            break;

        case EVENT_SCREEN_CHANGE:
//            rc = displayAdvanceScreen(pDisplay);
            iRequeueTime = 5000;
            break;

        case EVENT_SCREEN_REFRESH:
            rc = displayRefreshScreen(pDisplay);
            iRequeueTime = 500;
            break;

        case EVENT_LED_UPDATE:
        {
            /* Convenience */
            LPLED_INSTANCE pLEDInst = pDisplay->pLEDInstanceList[evt.iData];

            printf("Updating LED %d status\n", evt.iData);

            /* Get current state and adjust output accordingly */
            if (pLEDInst->bLEDOn || pLEDInst->freq == 0) {
                deviceSetLED(pDisplay->pDevice, evt.iData, &(pLEDInst->off));
            }
            else {
                deviceSetLED(pDisplay->pDevice, evt.iData, &(pLEDInst->on));
            }

            /* Calculate the requeue time */
            if (pLEDInst->freq == 0) {
                iRequeueTime = 2000; /* Frequency of zero means always off */
            }
            else {
                iRequeueTime = pLEDInst->bLEDOn ?
                    ((1000 * pLEDInst->dutyCycle) / pLEDInst->freq) :
                    ((1000 * (100 - pLEDInst->dutyCycle)) / pLEDInst->freq);
            }

            /* Toggle state and loop */
            pLEDInst->bLEDOn = ~(pLEDInst->bLEDOn);
            break;
        }

        case EVENT_POLL_LED_STATUS:
        {
            LPLED_INSTANCE pLEDInst = pDisplay->pLEDInstanceList[evt.iData];

            printf("Polling for LED %d status\n", evt.iData);

            /* Get LED data from module */
            (pLEDInst->pfn)(pLEDInst->context, pLEDInst->led_id,
                &(pLEDInst->on), &(pLEDInst->off),
                &(pLEDInst->dutyCycle), &(pLEDInst->freq));

            /* LEDs need only be updated every 2 seconds */
            iRequeueTime = 2000;
            break;

        }

        case EVENT_POLL_MODULE:
        {
            char modbuf[1024];
            //char * pBuf = modbuf;

            /* Get the text data too */
            /* moduleGetDisplayData(pModuleList[0], NULL, &pBuf); */
            sprintf(scrData, "%.20s", modbuf);

            /* Keep event in memory and requeue for 1 sec time */
            iRequeueTime = 1000;
            break;
        }

        case EVENT_TEXT_SCROLL:
        {
            char tmpBuf[1024];
            int strLen = -1;
            int dispWidth = 20;
            
            /* Amount of string to end */
            strLen = strlen(&longText[scrollPos]);

            if (strLen > dispWidth) {
                sprintf(tmpBuf, "%.*s", dispWidth, &longText[scrollPos]);
            }
            else {
                sprintf(tmpBuf, "%s %.*s", &longText[scrollPos], dispWidth - strLen - 1, longText);
            }

            /* Write to the display */
            deviceWriteText(pDisplay->pDevice, 0, 1, tmpBuf, strlen(tmpBuf));

            /* Scroll and check within bounds */
            scrollPos++;
            if (scrollPos >= strlen(longText)) {
                scrollPos = 0;
            }

            /* Keep event in memory and requeue for 50 msec time */
            iRequeueTime = 300;
            break;
        }

        case EVENT_KEY_PRESS:
            break;

        case EVENT_KEY_RELEASE:
            break;

        case EVENT_SET_TEMPERATURE_REPORTING:
            deviceSetTemperatureReporting(pDisplay->pDevice, (unsigned int) evt.iData);
            break;

        case EVENT_SET_FAN_REPORTING:
            deviceSetFanReporting(pDisplay->pDevice, (unsigned int) evt.iData);
            traceDebug(IILC_TRACE_CONTEXT, "Set fan reporting to %d", evt.iData);
            break;

        case EVENT_SET_FAN_STATE:
            deviceSetFanPower(pDisplay->pDevice, (unsigned int) evt.iData, 0, 0, 0);
            iRequeueTime = 3000;
            break;

        default:
            printf("Unknown event type received (%d)\n", evt.eType);
            break;
        }

        /* Should we requeue this event? */
        if (iRequeueTime >= 0) {
            eventQueueEnqueue(pEventQueue, &evt, iRequeueTime);
        }
    }

//end_of_function:

    IILC_TRACE_EXIT;
    return NULL;
}

void initCode()
{
    EVENT evt;
    LPEVENT_QUEUE pEventQueue = NULL;
    //int i = 0;

    /* ----- Initialise this display ----- */


    /* Create a display clear event */
    EVENT_INIT(evt, EVENT_SCREEN_CLEAR, NULL, 0);
    eventQueueEnqueue(pEventQueue, &evt, 0);

    /* Backlight on */
    EVENT_INIT(evt, EVENT_SET_BACKLIGHT, NULL, 40);
    eventQueueEnqueue(pEventQueue, &evt, 0);

    /* Contrast to something useful */
    EVENT_INIT(evt, EVENT_SET_CONTRAST, NULL, 90);
    eventQueueEnqueue(pEventQueue, &evt, 0);

    /* Start temp reporting */
    EVENT_INIT(evt, EVENT_SET_TEMPERATURE_REPORTING, NULL, 0x03);
    eventQueueEnqueue(pEventQueue, &evt, 0);

    /* Start fan reporting */
    EVENT_INIT(evt, EVENT_SET_FAN_REPORTING, NULL, 0x01);
    eventQueueEnqueue(pEventQueue, &evt, 0);

    /* Start fan at 100% */
    EVENT_INIT(evt, EVENT_SET_FAN_STATE, NULL, 100);
    eventQueueEnqueue(pEventQueue, &evt, 0);

    /* Repeatedly poll module 0 */
    EVENT_INIT(evt, EVENT_POLL_MODULE, NULL, 0);
    eventQueueEnqueue(pEventQueue, &evt, 0);

    /* Update the screen */
    EVENT_INIT(evt, EVENT_SCREEN_REFRESH, NULL, 0);
    eventQueueEnqueue(pEventQueue, &evt, 0);

    /* Scroll some text */
    EVENT_INIT(evt, EVENT_TEXT_SCROLL, NULL, 0);
    eventQueueEnqueue(pEventQueue, &evt, 0);




}


int displayGetSize(LPDISPLAY pDisplay, int * width, int * height)
{
    int rc = 0;

    *width = pDisplay->width;
    *height = pDisplay->height;

    return rc;
}

int displayScreenAdd(LPDISPLAY pDisplay, LPSCREEN pScreen)
{
    int rc = 0;
    LPSCREEN * pNewList = NULL;
    pthread_t * pNewThreadList = NULL;
    //pthread_t newTid;
    EVENT evt;

    /* Allocate a new list */
    pNewList = calloc(pDisplay->nScreenCount+1, sizeof(LPSCREEN));
    pNewThreadList = calloc(pDisplay->nScreenCount+1, sizeof(pthread_t));

    /* Do the copy */
    memcpy(pNewList, pDisplay->pScreenList, pDisplay->nScreenCount * sizeof(LPSCREEN));
    memcpy(pNewThreadList, pDisplay->screenThreads, pDisplay->nScreenCount * sizeof(pthread_t));

    /* Release old memory */
    free(pDisplay->pScreenList);
    free(pDisplay->screenThreads);

    /* Assign new data */
    pDisplay->pScreenList = pNewList;
    pDisplay->screenThreads = pNewThreadList;

    /* Add the extra one on the end */
    pNewList[pDisplay->nScreenCount] = pScreen;
    pthread_create(&(pNewThreadList[pDisplay->nScreenCount]), NULL, RenderThread, pScreen);

    /* Increase count */
    pDisplay->nScreenCount += 1;

    /* Remember to update us in the main Tx loop */
    EVENT_INIT(evt, EVENT_SCREEN_REFRESH, NULL, 0);
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

    return rc;
}



int displayRefreshScreen(LPDISPLAY pDisplay)
{
    int rc = 0;
    int i = 0;
    //int nWindows = 0;
    LPSCREEN pScreen = NULL;
    
    /* Buffer to draw to screen */
    char pszScreen[PHYSICAL_DEVICE_HEIGHT][PHYSICAL_DEVICE_WIDTH];

    /* Zap whole screen with spaces */
    memset(pszScreen, (int) ' ', sizeof(pszScreen));

    /* If we have no screens, let the user know */
    if (pDisplay->nScreenCount == 0) {
        char * pszNoConfig = "No screens setup";
        strncpy(pszScreen[0], pszNoConfig, strlen(pszNoConfig));
    }
    else {
        /* Get the current screen we are displaying */
        pScreen = pDisplay->pScreenList[pDisplay->nScreenCurrent];

        /* Render this line by line */
        for (i = 0; i < PHYSICAL_DEVICE_HEIGHT; i++) {
            screenDoRender(pScreen, i, pszScreen[i]);
        }
    }

    /* Write all data to the screen */
    for (i = 0; i < PHYSICAL_DEVICE_HEIGHT; i++) {
        rc = deviceWriteText(pDisplay->pDevice, 0, i, pszScreen[i], PHYSICAL_DEVICE_WIDTH);
    }

    return rc;
}

void displayCallbackTemp(void * context, unsigned int sensor, double degc)
{

}

void displayCallbackKeyEvent(void * context, PACKET_KEY_EVENT event)
{


}

void displayCallbackFan(void * context, unsigned int fan, unsigned int rpm)
{
    int i = 0;
    int nCount = 0;
    LPDISPLAY pDisplay = (LPDISPLAY) context;

    IILC_TRACE_ENTRY(IILC_MODULE, T_displayCallbackFan);
    traceDebug(IILC_TRACE_CONTEXT, "fan %u, rpm %u", fan, rpm);

    /* Iterate through all callbacks */
    vectorGetSize(pDisplay->pFanListeners, &nCount);
    for (i = 0; i < nCount; i++) {

        LPFAN_LISTENER pListener;

        /* Get the next callback */
        vectorGetElementAt(pDisplay->pFanListeners, i, (void **) &pListener);

        /* Is this callback relevant for this fan? */
        if (pListener->fan_id != fan) {
            continue; /* No - skip to next callback */
        }

        /* Yes - perform the callback */
        (pListener->pfn)(pListener->context, pListener->reportAs, rpm);
    }

    IILC_TRACE_EXIT;
}

