/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#if defined(_WIN32)
#   include "pthread_w32.h"
#else
#   include <pthread.h>
#   include <errno.h>
#endif

#include "lcd.h"
#include "config.h"
#include "trace.h"
#include "vector.h"
#include "device.h"
#include "module.h"
#include "display.h"
#include "protos.h"


#define IILC_MODULE MODULE_TYPE_APP


/****************************************************************************
 *                                                                          *
 * Parser-related goodness                                                  *
 *                                                                          *
 ****************************************************************************/

/*
 * A flag to set to switch on flex debug mode
 */
extern int yydebug;

/*
 * The file read by the parser
 */
extern FILE * yyin;

/*
 * Key under which to find the current parser in TLS
 */
static pthread_key_t g_keyParserThis;

/**
  * Variable which holds ensures parser thread data is initialised once only.
  */
static pthread_once_t parse_init = PTHREAD_ONCE_INIT;

/**
  * Macro which ensures the parser is initialised once and once only
  */
#define PARSER_INITIALISE  (void) pthread_once(&parse_init, parserInitialise);


/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/


typedef struct _tag_config_object
{
    enum displayObjectType type;

    /** Internal ID as reported to the physical device */
    unsigned int physical_id;

    /** Name as used in the config file */
    char * pszName;

    /** The module which is allowed to control this object */
    LPCONFIG_MODULE pControllingModule;

    /** The ID which the module will use to control this object */
    unsigned int nModuleControlAsID;

    /** A vector containing modules who wish to receive notifications on this object */
    LPVECTOR pReportModulesList;

    /** A vector of 'int' that shows the ID against which modules should receive reports */
    LPVECTOR pReportIDList;

} CONFIG_OBJECT;



typedef struct _tag_config_device
{
    /* ----- Child data ----- */

    /** A vector of LPCONFIG_OBJECT elements listing resources available */
    LPVECTOR pObjectList;

    /* ----- Config data ----- */

    char * pszName;

    char * pszDeviceClass;

    char * pszPort;
    
    /* ----- Runtime data ----- */

    LPDEVICE pDevice;

} CONFIG_DEVICE;



typedef struct _tag_config_module
{
    /** Name of the library to load */
    char * pszName;

    /** The actual module handle in use */
    LPMODULE pModule;

} CONFIG_MODULE;


typedef struct _tag_config_display
{
    /* ----- Child data ----- */

    /** A display is run by mutiple modules */
    LPVECTOR pModules;

    /** Each display can show multiple screens */
    LPVECTOR pScreens;

    /* ----- Config data ----- */

    /** The single device we are controlling */
    LPCONFIG_DEVICE pConfigDevice;

    /** Name of this display */
    char * pszName;
    
    /* ----- Runtime data ----- */

    /** Pointer to the display to use */
    LPDISPLAY pDisplay;

} CONFIG_DISPLAY;


typedef struct _tag_config_app
{
    /* ----- Child data ----- */

    /** The list of displays this application is co-ordinating */
    LPVECTOR pDisplays;

    /* ----- Config data ----- */


    /* ----- Runtime data ----- */


} CONFIG_APP;



/****************************************************************************
 *                                                                          *
 * Parse-time functions                                                     *
 *                                                                          *
 ****************************************************************************/


void parserInitialise(void)
{
    /* Initialise the stack depth TLS key */
    (void) pthread_key_create(&g_keyParserThis, NULL);
}

LPCONFIG_APP parserGetCurrentConfig()
{
    LPCONFIG_APP pInst = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_parserGetCurrentConfig);

    PARSER_INITIALISE;

    /* Grab the current parser on this thread */
    pInst = (LPCONFIG_APP) pthread_getspecific(g_keyParserThis);
    if (pInst == NULL) {
        printf("Failed to obtain parser on current thread\n");
        exit(1);
    }

    traceDebug(IILC_TRACE_CONTEXT, "pInst %p", pInst);
    IILC_TRACE_EXIT;
    return pInst;
}

LPCONFIG_DEVICE parserGetCurrentDevice()
{
    LPCONFIG_DEVICE pDevice = NULL;
    LPCONFIG_DISPLAY pDisplay = NULL;
    
    pDisplay = parserGetCurrentDisplay();
    pDevice = pDisplay->pConfigDevice;

    return pDevice;
}

LPCONFIG_DISPLAY parserGetCurrentDisplay()
{
    LPCONFIG_DISPLAY pDisplay = NULL;
    LPCONFIG_APP pInst = NULL;
    int size = 0;

    pInst = parserGetCurrentConfig();

    /* Get the last element in the vector */
    vectorGetSize(pInst->pDisplays, &size);
    vectorGetElementAt(pInst->pDisplays, size-1, (void **) &pDisplay);

    return pDisplay;
}

LPCONFIG_MODULE parserModuleGetCurrent()
{
    LPCONFIG_DISPLAY pDisplay = NULL;
    LPCONFIG_MODULE pModule = NULL;
    int size = 0;

    pDisplay = parserGetCurrentDisplay();

    /* Get the last element in the vector */
    vectorGetSize(pDisplay->pModules, &size);
    vectorGetElementAt(pDisplay->pModules, size-1, (void **) &pModule);

    return pModule;
}


int parserParse(const char * pszConfigFile, LPCONFIG_APP * ppAppConfig)
{
    int rc = 0;
    LPCONFIG_APP pConfig = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_parserParse);
    traceDebug(IILC_TRACE_CONTEXT,
        "pszConfigFile %p, ppAppConfig %p", pszConfigFile, ppAppConfig);
        
    PARSER_INITIALISE;

    /* Create a config instance */
    pConfig = (LPCONFIG_APP) calloc(1, sizeof(CONFIG_APP));
    if (pConfig == NULL) {
        rc = ERR_MEM_ALLOC_FAILED;
        goto end_of_function;
    }

    /* Make this available on the current thread */
    pthread_setspecific(g_keyParserThis, (void *) pConfig);

    /* Init list of displays */
    vectorCreate(&(pConfig->pDisplays));

    /* Open the file */
    yyin = fopen(pszConfigFile, "r");
    yydebug = 0;
    
    /* Parse the file */
    startParse();

    /* Close the file */
    fclose(yyin);

    /* Drop the TLS data */
    pthread_setspecific(g_keyParserThis, NULL);

    /* Return config to user */
    if (ppAppConfig != NULL) {
        *ppAppConfig = pConfig;
    }

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}


int configInstanceDestroy(LPCONFIG_APP * ppInst)
{
    LPCONFIG_APP pInst = *ppInst;

    free(pInst);

    *ppInst = NULL;

    return 0;
}

void parserDeviceBegin(char * pszDevice)
{
    int rc = 0;

    LPCONFIG_DEVICE pNewDevice = NULL;
    LPCONFIG_DISPLAY pDisplay = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_parserBeginDevice);
    traceDebug(IILC_TRACE_CONTEXT, "pszDevice %p", pszDevice);

    /* Get the currently configured display */
    pDisplay = parserGetCurrentDisplay();

    /* Syntax check for config file */
    if (pDisplay->pConfigDevice != NULL) {
        traceError(IILC_TRACE_CONTEXT, "A device has already been specified for this display");
        goto end_of_function;
    }

    /* Create new device & init */
    pNewDevice = (LPCONFIG_DEVICE) malloc(sizeof(CONFIG_DEVICE));
    pNewDevice->pszName        = pszDevice;
    pNewDevice->pszDeviceClass = NULL;
    pNewDevice->pObjectList    = NULL;
    
    /* Create the object list vector */
    rc = vectorCreate(&(pNewDevice->pObjectList));
    if (rc) {
        traceError(IILC_TRACE_CONTEXT, "Failed to create object vector");
        goto end_of_function;
    }

    /* Add new data into current display */
    pDisplay->pConfigDevice = pNewDevice;

end_of_function:

    IILC_TRACE_EXIT;
}


void parserDeviceEnd()
{
    IILC_TRACE_ENTRY(IILC_MODULE, T_parserEndDevice);

    IILC_TRACE_EXIT;
}


void parserDeviceUseFan(int id, char * pszName)
{
    int rc = 0;

    LPCONFIG_DEVICE pDevice = NULL;
    LPCONFIG_OBJECT pObject = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_parserDeviceUseFan);
    traceDebug(IILC_TRACE_CONTEXT, "id %d, pszName %s", id, pszName);

    pDevice = parserGetCurrentDevice();

    /* Check this name hasn't been declared already */
    rc = parserDeviceFindObject(pDevice, pszName, &pObject);
    if (rc != ERR_OBJECT_NOT_FOUND) {
        traceDebug(IILC_TRACE_CONTEXT, 
            "The object %s has already been declared on this device");
        rc = ERR_INVALID_CONFIG;
        goto end_of_function;
    }

    /* Good - create this new object */
    rc = parserDeviceCreateObject(id, pszName, OBJECT_TYPE_FAN, &pObject);
    if (rc) {
        traceError(IILC_TRACE_CONTEXT, "Error creating fan entry");
        goto end_of_function;
    }

    /* Add to the device vector*/
    vectorAdd(pDevice->pObjectList, pObject);

end_of_function:

    IILC_TRACE_EXIT;
}

int parserDeviceCreateObject(int physical_id, char * pszName, 
                             enum displayObjectType type, LPCONFIG_OBJECT * ppObject)
{
    int rc = 0;
    LPCONFIG_OBJECT pObject = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_parserDeviceCreateObject);

    /* Alloc and init the structure */
    pObject = (LPCONFIG_OBJECT) malloc(sizeof(CONFIG_OBJECT));
    if (pObject == NULL) {
        rc = ERR_MEM_ALLOC_FAILED;
        goto end_of_function;
    }

    /* init */
    pObject->physical_id        = physical_id;
    pObject->pszName            = pszName;
    pObject->type               = type;
    pObject->nModuleControlAsID = 0;
    pObject->pControllingModule = NULL;
    pObject->pReportIDList      = NULL;
    pObject->pReportModulesList = NULL;

    /* Vector inits */
    rc = vectorCreate(&(pObject->pReportIDList));
    if (rc) { goto end_of_function; }

    rc = vectorCreate(&(pObject->pReportModulesList));
    if (rc) { goto end_of_function; }

    /* Return to the caller */
    *ppObject = pObject;

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return (rc);
}

void parserDeviceUseLED(int id, char * pszName)
{



}

void parserDeviceUseTemp(int id, char * pszName)
{




}


void parserSetDisplayDevice(char * pszDevice)
{
}

void parserSetModuleLibrary(char * pszLibrary)
{
    LPCONFIG_MODULE pModule = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_parserSetModuleLibrary);
    traceDebug(IILC_TRACE_CONTEXT, "pszLibrary %s", pszLibrary);

    /* Get the current module */
    pModule = parserModuleGetCurrent();

    /* Set the library name */
    pModule->pszName = pszLibrary;

    IILC_TRACE_EXIT;
}


int parserDeviceFindObject(LPCONFIG_DEVICE pDevice, char * pszObject, LPCONFIG_OBJECT * ppObject)
{
    int i      = 0;
    int rc     = 0;
    int nCount = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_parserDeviceFindObject);
    traceDebug(IILC_TRACE_CONTEXT, 
        "pDevice %p, pszObject %s, ppObject %p", pDevice, pszObject, ppObject);

    /* Check each item in the list */
    vectorGetSize(pDevice->pObjectList, &nCount);
    for (i = 0; i < nCount; i++) {

        LPCONFIG_OBJECT pObject = NULL;
        rc = vectorGetElementAt(pDevice->pObjectList, i, (void **) &pObject);
        if (strcmp(pObject->pszName, pszObject) == 0) {

            /* Return to the user */
            traceDebug(IILC_TRACE_CONTEXT, "Found the object named [%s]", pszObject);
            *ppObject = pObject;
            goto end_of_function;
        }
    }

    /* Not found - exit with error */
    rc = ERR_OBJECT_NOT_FOUND;

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return (rc);
}

void parserModuleControlObject(char * pszObject, int report_as_id)
{
    int rc     = 0;

    LPCONFIG_MODULE  pModule  = NULL;
    LPCONFIG_DISPLAY pDisplay = NULL;
    LPCONFIG_DEVICE  pDevice  = NULL;
    LPCONFIG_OBJECT  pObject  = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_parserModuleControlObject);
    traceDebug(IILC_TRACE_CONTEXT, "pszObject %s, report_as_id %d", pszObject, report_as_id);

    /* Get the current module */
    pModule = parserModuleGetCurrent();

    /* Lookup the object requested from the device used */
    pDisplay = parserGetCurrentDisplay();

    /* Check the device has been configured */
    pDevice = pDisplay->pConfigDevice;
    if (pDevice == NULL) {
        traceError(IILC_TRACE_CONTEXT, "No device configured for this display");
        goto end_of_function;
    }

    /* Find the object */
    rc = parserDeviceFindObject(pDevice, pszObject, &pObject);
    if (rc) {
        traceError(IILC_TRACE_CONTEXT, "Could not find the object named %s", pszObject);
        goto end_of_function;
    }

    /* Let this object know the module which is controlling it */
    pObject->pControllingModule = pModule;

end_of_function:

    IILC_TRACE_EXIT;
}

void parserModuleReportObject(char * pszObject, int report_as_id)
{
    LPCONFIG_MODULE pModule;
    LPCONFIG_DEVICE pDevice;
    LPCONFIG_OBJECT pObject = NULL;
    int * pReportId = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_parserModuleReportObject);
    traceDebug(IILC_TRACE_CONTEXT, 
        "pszObject %s, report_as_id %d", pszObject, report_as_id);

    /* Get the current module being configured */
    pModule = parserModuleGetCurrent();

    /* Get the device this will refer to */
    pDevice = parserGetCurrentDevice();

    /* Lookup the object being referred to */
    parserDeviceFindObject(pDevice, pszObject, &pObject);
    if (pObject == NULL) {
        traceError(IILC_TRACE_CONTEXT,
            "The object [%s] was not found in device [%s]", pszObject, pDevice->pszName);
        goto end_of_function;
    }

    /* Add this module to the list of reports the device should make */
    pReportId = (int *) malloc(sizeof(int));
    *pReportId = report_as_id;

    vectorAdd(pObject->pReportIDList, (void *) pReportId);
    vectorAdd(pObject->pReportModulesList, pModule);

end_of_function:

    IILC_TRACE_EXIT;
}

void parserModuleVirtualObject(int physical_id, char * pszObject)
{
    IILC_TRACE_ENTRY(IILC_MODULE, T_parserModuleVirtualObject);
    traceDebug(IILC_TRACE_CONTEXT, "physical_id %d, pszObject %s", physical_id, pszObject);



    IILC_TRACE_EXIT;
}

void parserSetDeviceClass(char * pszClass)
{
    parserGetCurrentDevice()->pszDeviceClass = pszClass;
}


void parserSetDevicePort(char * pszPort)
{
    parserGetCurrentDevice()->pszPort = pszPort;
}


void parserBeginDisplay(char * pszDisplay)
{
    LPCONFIG_DISPLAY pNewDisplay = NULL;
    LPCONFIG_APP pConfig = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_parserBeginDisplay);
    traceDebug(IILC_TRACE_CONTEXT, "pszDisplay %s", pszDisplay);

    /* Get current configuration */
    pConfig = parserGetCurrentConfig();

    /* Init the struct */
    pNewDisplay = (LPCONFIG_DISPLAY) malloc(sizeof(CONFIG_DISPLAY));
    pNewDisplay->pConfigDevice = NULL;
    pNewDisplay->pDisplay      = NULL;
    pNewDisplay->pModules      = NULL;
    pNewDisplay->pScreens      = NULL;
    pNewDisplay->pszName       = pszDisplay;

    /* Init the vectors */
    vectorCreate(&(pNewDisplay->pModules));
    vectorCreate(&(pNewDisplay->pScreens));

    /* Insert into the instance vector */
    vectorAdd(pConfig->pDisplays, (void *) pNewDisplay);

    IILC_TRACE_EXIT;
}

void parserEndDisplay()
{
    IILC_TRACE_ENTRY(IILC_MODULE, T_parserEndDisplay);

    IILC_TRACE_EXIT;
}

void parserBeginModule(char * pszModule)
{
    LPCONFIG_MODULE pNewModule = NULL;
    LPCONFIG_DISPLAY pDisplay = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_parserBeginModule);
    traceDebug(IILC_TRACE_CONTEXT, "pszModule %s", pszModule);

    /* Get the display we are currently configuring */
    pDisplay = parserGetCurrentDisplay();

    /* Init the struct */
    pNewModule = (LPCONFIG_MODULE) malloc(sizeof(CONFIG_MODULE));
    pNewModule->pModule = NULL;
    pNewModule->pszName = pszModule;

    /* Insert into the correct instance vector */
    vectorAdd(pDisplay->pModules, (void *) pNewModule);

    IILC_TRACE_EXIT;
}

void parserEndModule()
{
    IILC_TRACE_ENTRY(IILC_MODULE, T_parserEndModule);

    IILC_TRACE_EXIT;
}

void parserBeginScreen(char * pszScreen)
{
    printf("Begin screen %s\n", pszScreen);
}

void parserEndScreen()
{
    printf("End screen\n");
}


int configGetDeviceName(LPCONFIG_DEVICE pDevice, const char ** ppszName)
{
    int rc = 0;



    return rc;
}


int configGetDeviceClass(LPCONFIG_DEVICE pDevice, const char ** ppszClass)
{
    int rc = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_configGetDeviceClass);
    traceDebug(IILC_TRACE_CONTEXT, "pDevice %p, ppszClass %p", pDevice, ppszClass);

    /* Store for user */
    *ppszClass = pDevice->pszDeviceClass;
    traceDebug(IILC_TRACE_CONTEXT, "Device class = [%s]", *ppszClass);

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}


int configGetDevicePort(LPCONFIG_DEVICE pDevice, const char ** ppszPort)
{
    int rc = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_configGetDevicePort);
    traceDebug(IILC_TRACE_CONTEXT, "pDevice %p, ppszPort %p", pDevice, ppszPort);

    /* Store for user */
    *ppszPort = pDevice->pszPort;
    traceDebug(IILC_TRACE_CONTEXT, "Port = [%s]", *ppszPort);

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}

int configGetDisplayDevice(LPCONFIG_APP pConfig, int index, int * pnDeviceIndex)
{
    int rc = 0;


    return rc;
}

int configGetDisplayDescr(LPCONFIG_DISPLAY pDisplay, const char ** ppszDescr)
{
    int rc = 0;

    //*ppszDescr = pConfig->device.pDefs[index]->pszName;

    return rc;
}


int configGetDisplayCount(LPCONFIG_APP pConfig, int * pnCount)
{
    int rc = 0;
    int nCount = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_configGetDisplayCount);
    traceDebug(IILC_TRACE_CONTEXT, "pConfig %p, pnCount %p", pConfig, pnCount);

    if (pnCount == NULL) {
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    rc = vectorGetSize(pConfig->pDisplays, &nCount);
    if (rc) {
        goto end_of_function;
    }

    *pnCount = nCount;
    traceDebug(IILC_TRACE_CONTEXT, "*pnCount %d", *pnCount);

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}




int configGetModuleName(LPCONFIG_DISPLAY pDisplay, int nModuleIndex, char ** ppszName)
{
    int rc = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_configGetModuleName);
    traceDebug(IILC_TRACE_CONTEXT,
        "pDisplay %p, nModuleIndex %d, ppszName %p", pDisplay, nModuleIndex, ppszName);

    if (pDisplay == NULL) {
        traceError(IILC_TRACE_CONTEXT, "pDisplay == NULL");
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    if (ppszName == NULL) {
        traceError(IILC_TRACE_CONTEXT, "ppszName == NULL");
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

//    *ppszName = pDisplay->module.pDefs[nModuleIndex]->pszName;

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}


int configGetDeviceCount(LPCONFIG_APP pConfig, int * pnCount)
{
    int rc = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_configGetDeviceCount);
    traceDebug(IILC_TRACE_CONTEXT, "pConfig %p, pnCount %p", pConfig, pnCount);

    //*pnCount = pConfig->device.nSize;

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}


int configGetDisplayDeviceIndex(LPCONFIG_APP pConfig, int nDisplayIndex, int * nDeviceIndex)
{



    return 0;
}


int processConfigDevice(LPCONFIG_DEVICE pDevice)
{
    int rc = 0;

    const char * pszDescr       = NULL;
    const char * pszDevicePort  = NULL;
    const char * pszDeviceClass = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_processConfigDevice);
    traceDebug(IILC_TRACE_CONTEXT, "pDevice %p", pDevice);

    /* Get the name of this device (always present) */
    rc = configGetDeviceName(pDevice, &pszDescr);
    if (rc) {
        traceError(IILC_TRACE_CONTEXT, "Could not get name for device");
        goto end_of_function;
    }

    /* Get the device config */
    rc = configGetDeviceClass(pDevice, &pszDeviceClass);
    if (rc) {
        traceError(IILC_TRACE_CONTEXT, "Could not get device class for device [%s]", pszDescr);
        goto end_of_function;
    }

    /* Get the port it is connected to */
    rc = configGetDevicePort(pDevice, &pszDevicePort);
    if (rc) {
        traceError(IILC_TRACE_CONTEXT, "Could not get port for device [%s]", pszDescr);
        goto end_of_function;
    }

    /* Create a new instance of the device */
    rc = deviceCreate(pszDeviceClass, pszDevicePort, &(pDevice->pDevice));
    if (rc) {
        traceError(IILC_TRACE_CONTEXT, "Could not create an instance of device [%s]", pszDescr);
        goto end_of_function;
    }

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}


int processConfigDisplay(LPCONFIG_DISPLAY pDisplay)
{
    int rc     = 0;
    int i = 0, j = 0;
    int nCount = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_processConfigDisplay);
    traceDebug(IILC_TRACE_CONTEXT, "pDisplay %p", pDisplay);


    /* ----- Init the device used ----- */

    if (pDisplay->pConfigDevice == NULL) {
        traceError(IILC_TRACE_CONTEXT,
            "No device configured for display \"%s\"", pDisplay->pszName);
        goto end_of_function;
    }

    /* Create the device */
    rc = processConfigDevice(pDisplay->pConfigDevice);
    if (rc) {
        traceError(IILC_TRACE_CONTEXT, "Error creating device");
        goto end_of_function;
    }

    /* ----- Create the display ----- */

    rc = displayCreate(pDisplay->pConfigDevice->pDevice, &(pDisplay->pDisplay));
    if (rc) {
        traceError(IILC_TRACE_CONTEXT, "Error creating display");
        goto end_of_function;
    }

    traceDebug(IILC_TRACE_CONTEXT, "Configuring display [%s]", pDisplay->pszName);

    /* ----- Init the modules list ----- */

    /* Get the number of modules requested */
    rc = vectorGetSize(pDisplay->pModules, &nCount);
    if (rc) {
        traceError(IILC_TRACE_CONTEXT,
            "Could not get the number of modules for display %s", pDisplay->pszName);
        goto end_of_function;
    }

    /* Loop through each module */
    for (i = 0; (i < nCount) && (rc == 0); i++) {

        LPCONFIG_MODULE pModule = NULL;
        rc = vectorGetElementAt(pDisplay->pModules, i, (void **) &pModule);
        if (!rc) {
            rc = processConfigModule(pModule);
        }
    }

    /* ----- Setup the callbacks ----- */

    /* Get all the objects connected to the device */
    vectorGetSize(pDisplay->pConfigDevice->pObjectList, &nCount);
    for (i = 0; i < nCount; i++) {

        int nReports;
        LPCONFIG_OBJECT pObject = NULL;

        vectorGetElementAt(pDisplay->pConfigDevice->pObjectList, i, (void **) &pObject);

        /* Does this have any modules wishing to have reports on it? */
        vectorGetSize(pObject->pReportModulesList, &nReports);
        traceDebug(IILC_TRACE_CONTEXT, "size %d", nReports);
        for (j = 0; j < nReports; j++) {

            LPCONFIG_MODULE pModule;
            int * pnReportID;

            /* Get the module in question & the ID to report as */
            vectorGetElementAt(pObject->pReportModulesList, j, (void **) &pModule);
            vectorGetElementAt(pObject->pReportIDList, j, (void **) &pnReportID);
            
            switch (pObject->type) {
            case OBJECT_TYPE_FAN:
            {
                /* Let the display know we wish to receive reports for this */
                displayAddFanReportListener(pDisplay->pDisplay, 
                    pObject->physical_id, moduleFanSpeedReport, pModule->pModule, *pnReportID);
                break;
            }
            case OBJECT_TYPE_TEMP:
            {
                break;
            }
            default:
                /* Just ignore any reports from other objects */
                break;
            }
        }

    }

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}


int processConfigModule(LPCONFIG_MODULE pModule)
{
    int rc = 0;

    /* Initialise the module */
    rc = moduleInit(pModule->pszName, &(pModule->pModule));

    return rc;
}

int processConfigFile(LPCONFIG_APP pConfig)
{
    int rc = 0;
    int i = 0;
    int count = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_processConfigFile);
    traceDebug(IILC_TRACE_CONTEXT, "pConfig %p", pConfig);

    /* Number of displays in config file */
    rc = configGetDisplayCount(pConfig, &count);
    if (rc) {
        goto end_of_function;
    }

    if (count == 0) {
        rc = ERR_INVALID_CONFIG;
        traceError(IILC_TRACE_CONTEXT, "No displays configured for application. Exiting");
        goto end_of_function;
    }

    /* Initialise each one of them */
    for (i = 0; (i < count) && (rc == 0); i++) {

        LPCONFIG_DISPLAY pDisplay = NULL;

        rc = vectorGetElementAt(pConfig->pDisplays, i, (void **) &pDisplay);
        if (!rc) {
            rc = processConfigDisplay(pDisplay);
        }
    }

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}


int main(int argc, char ** argv)
{
    /* One configuration per application instance */
    LPCONFIG_APP pConfig = NULL;
    
    /* Handy loop variable */
    int i = 0;

    /* Generic return code */
    int rc = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_main);

    /* Process the config file */
    parserParse("../conf/lcd.conf", &pConfig);

    if (argc > 2) {
        exit(0);
    }

    /* Examine the config instance */
    processConfigFile(pConfig);

    if (argc > 1) {
        exit(0);
    }

    /* Start each of the display instances */
    {
        int nCount = 0;
        LPCONFIG_DISPLAY pDisplay = NULL;

        rc = vectorGetSize(pConfig->pDisplays, &nCount);
        if (rc) {
            traceError(IILC_TRACE_CONTEXT, "Could not determine number of displays");
            goto end_of_function;
        }

        for (i = 0; (i < nCount) && (rc == 0); i++) {
            rc = vectorGetElementAt(pConfig->pDisplays, i, (void **) &pDisplay);
            if (!rc) {
                rc = displayStart(pDisplay->pDisplay);
            }
        }
    }

    if (!rc) {
        /* No error - so sleep this thread for a bit */
        SLEEP(10000);
    }

    /* Stop each of the display instances */
    {
        int nCount = 0;
        LPCONFIG_DISPLAY pDisplay = NULL;

        rc = vectorGetSize(pConfig->pDisplays, &nCount);
        if (rc) {
            traceError(IILC_TRACE_CONTEXT, "Could not determine number of displays");
            goto end_of_function;
        }

        for (i = 0; (i < nCount) && (rc == 0); i++) {
            rc = vectorGetElementAt(pConfig->pDisplays, i, (void **) &pDisplay);
            if (!rc) {
                rc = displayStop(pDisplay->pDisplay);
            }
        }
    }

end_of_function:

    /* Delete the config instance */
    if (pConfig != NULL) {
        configInstanceDestroy(&pConfig);
    }

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}

//
//    /* Work out how many modules are required */
//    pInstance->moduleCount = sizeof(pszModuleList) / sizeof(char *);
//
//    /* Init list of modules */
//    pInstance->pModuleList = malloc(pInstance->moduleCount * sizeof(LPMODULE));
//    if (pInstance->pModuleList == NULL) {
//        traceError(IILC_TRACE_CONTEXT, "Failed to malloc module list");
//        goto end_of_function;
//    }
//    
//    /* Load the modules */
//    for (i = 0; i < pInstance->moduleCount; i++) {
//        rc = moduleInit(pszModuleList[i], &(pInstance->pModuleList[i]));
//        if (rc) {
//            printf("Error loading module %s\n", pszModuleList[i]);
//            goto end_of_function;
//        }
//    }
//
//    /* Create a display we can talk to */
//    displayCreate(&(pInstance->pDisplay));
//
//    /* Convenience */
//    pDisplay = pInstance->pDisplay;
//
//    /* Just one screen for now */
//    screenCreate(&pScreen);
//
//    /* Add this screen to the display */
//    displayScreenAdd(pDisplay, pScreen);
//
//    /* ----- Initialise config of the LED module ----- */
//    pModule = pInstance->pModuleList[MODULE_ID_LEDS];
//
//    /* Set LEDs 0 & 1 to be controlled by this module */
//    //displaySetLEDController(pDisplay, 0, moduleGetLEDStatus, pModule, 0);
//    //displaySetLEDController(pDisplay, 1, moduleGetLEDStatus, pModule, 1);
//
//
//    /* ----- Init config of the fans module ----- */
//    pModule = pInstance->pModuleList[MODULE_ID_FANS];
//
//    /* This module wishes to receive reports on fan 0 status (report as fan 1) */
//    displayAddFanReportListener(pDisplay, 0, moduleFanSpeedReport, pModule, 1);
//    displayAddFanReportListener(pDisplay, 1, moduleFanSpeedReport, pModule, 2);
//
//    /* Control the fans based on temperature reports */
//    //displayAddTempReportListener(pDisplay, 0, moduleTemperatureReport, pModule, 0);
//    //displayAddTempReportListener(pDisplay, 1, moduleTemperatureReport, pModule, 1);
//
//    /* This module controls all of the fans on this display */
//    //displaySetFanController(pDisplay, 0, moduleGetFanPower, pModule, 0);
//    //displaySetFanController(pDisplay, 1, moduleGetFanPower, pModule, 1);
//    //displaySetFanController(pDisplay, 2, moduleGetFanPower, pModule, 2);
//    //displaySetFanController(pDisplay, 3, moduleGetFanPower, pModule, 3);
//
//    /* Set LED 2 to be controlled by this module (report to callback as LED 0) */
//    //displaySetLEDController(pDisplay, 2, moduleGetLEDStatus, pModule, 0);
//
//    /* Place a fans window at (0,0) to be 20x2 with text data */
//    screenAddWindowText(pScreen, 0, 0, 20, 2, moduleGetDisplayData, pModule);
//
//
//    /* ----- Init config of the temp module ----- */
//    pModule = pInstance->pModuleList[MODULE_ID_TEMP];
//
//    /* This module wants to receive reports on temperatures from sensors 0 & 1 */
//    //displayAddTempReportListener(pDisplay, 0, moduleTemperatureReport, pModule, 0);
//    //displayAddTempReportListener(pDisplay, 1, moduleTemperatureReport, pModule, 1);
//
//    /* Place a temp window at (0, 2) to be 20x2 with text data */
//    screenAddWindowText(pScreen, 0, 2, 20, 2, moduleGetDisplayData, pModule);
//
//    /* ----- Init config of the overall controlling app ----- */
//
//    /* We also wish to receive key press / release notifications */
//    // displayAddKeyEventListener(pDisplay, keypressCallback, pInstance);
//
//
//
//    /* ---- Display now initialised and ready to start ----- */
//
//    
//    /* Start up the display */
//    rc = displayStart(pDisplay);
//    if (rc) { goto end_of_function; }
//
//
//    SLEEP(5000);
//
//    /* ----- Display now running on its own ----- */
//
//
//    /* Wait until the main thread terminates */
////    pthread_join(pInstance->tid, NULL);
//
//
//    /* ---- Main loop has shut down ----- */
//
//    /* Shutdown the display */
//    rc = displayStop(pDisplay);
//
//
//
//    /* Destroy the display instance */
//    rc = displayDestroy(&(pDisplay));


    ///* Unload the modules */
    //for (i = 0; i < pInstance->moduleCount; i++) {
    //    if (pInstance->pModuleList[i] != NULL) {
    //        moduleUnload(&(pInstance->pModuleList[i]));
    //    }
    //}

    ///* Clear module list memory */
    //if (pInstance->pModuleList != NULL) {
    //    free(pInstance->pModuleList);
    //}



int keypressCallback(void * context, PACKET_KEY_EVENT key_event)
{
    switch (key_event) {

    case KEY_UP_PRESS:
    case KEY_UP_RELEASE:
    case KEY_DOWN_PRESS:
    case KEY_DOWN_RELEASE:
    case KEY_LEFT_PRESS:
    case KEY_LEFT_RELEASE:
    case KEY_RIGHT_PRESS:
    case KEY_RIGHT_RELEASE:
    case KEY_ENTER_PRESS:
    case KEY_ENTER_RELEASE:
    case KEY_EXIT_PRESS:
    case KEY_EXIT_RELEASE:
        break;

    } /* switch */

    return 0;
}

void CallbackFanReport(void * context, unsigned int fan, unsigned int rpm)
{



}

