/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#include <stdio.h>
#include <stdarg.h>

#if defined(_WIN32)
#   include "pthread_w32.h"
#   define vsnprintf _vsnprintf
#else
#   include <pthread.h>
#endif

#include "trace.h"


/**
  * Main trace routine, only used internally.
  */

static void traceData(traceContextData * pContext,
                      char symbol,
                      char * fmt,
                      va_list args);

/**
  * Variable which holds ensures trace is initialised once only.
  */
static pthread_once_t trace_init = PTHREAD_ONCE_INIT;

/**
  * Macro which allows a simple means of modifying the trace init routine.
  */

#define TRACE_INITIALISE  (void) pthread_once(&trace_init, traceInitialise);

/**
  * The file descriptor to write trace to.
  */
static FILE * g_file = NULL;

/*
 * State flags for each type of trace function
 */

static unsigned char g_bFuncEnabled = 1; /* Function entry / exit trace   */
static unsigned char g_bSev1Enabled = 1; /* Really low-level debugging    */
static unsigned char g_bSev2Enabled = 1; /* Extra info about goings-on    */
static unsigned char g_bSev3Enabled = 1; /* Warnings about stuff          */
static unsigned char g_bSev4Enabled = 1; /* Internally-recoverable errors */
static unsigned char g_bSev5Enabled = 1; /* Catastrophic errors           */

/*
 * Key under which to find the stack depth in TLS
 */

static pthread_key_t g_keyStackDepth;



/********************************************************************************
 *      Internal functions                                                      *
 ********************************************************************************/


void traceInitialise(void)
{
    /* Initialise the stack depth TLS key */
    (void) pthread_key_create(&g_keyStackDepth, NULL);

    /* Open the file for writing */
    /* g_file = fopen("trace.txt", "a"); */
}

#define TRACE_BUFLEN 1024

#define FIFTY_SPACES  "                                                  "

void traceData(traceContextData * pContext, char symbol, char * fmt, va_list args)
{
    char    message[TRACE_BUFLEN];
    int   * pStackDepth = NULL;
    int     nStackDepth = 0;
    int     nIndent     = 0;

    /* Grab the current stack depth */
    pStackDepth = (int *) pthread_getspecific(g_keyStackDepth);
    if (pStackDepth == NULL) {
        pStackDepth = (int *) malloc(sizeof(int));
        *pStackDepth = 0;
        printf("Stack depth reset\n");
        pthread_setspecific(g_keyStackDepth, (void *) pStackDepth);
    }

    /* Get the current depth for this thread */
    nStackDepth = *pStackDepth;

    if (symbol == '<') { nStackDepth--; }

    /* Calculate the indent required */
    nIndent = nStackDepth * 2;

    /* Format the message */
    sprintf(message, "%c %*s", symbol, nIndent, FIFTY_SPACES);
    vsnprintf(&message[3+nIndent], TRACE_BUFLEN-3-nIndent, fmt, args);
    printf("%s\n",message);

    if (symbol == '>') { nStackDepth++; }

    /* Update the stack depth on exit */
    *pStackDepth = nStackDepth;
}



/********************************************************************************
 *      Exported functions                                                      *
 ********************************************************************************/


void traceEntry(traceContextData * pContext)
{
    TRACE_INITIALISE;

    if (g_bFuncEnabled) {
        traceData(pContext, '>', &(pContext->pszFunction[2]), NULL);
    }
}


void traceExit(traceContextData * pContext)
{
    TRACE_INITIALISE;

    if (g_bFuncEnabled) {
        traceData(pContext, '<', &(pContext->pszFunction[2]), NULL);
    }
}

void traceDebug(traceContextData * pContext, char * fmt, ...)
{
    va_list args;

    TRACE_INITIALISE;

    if (g_bSev1Enabled) {
        va_start(args, fmt);
        traceData(pContext, 'D', fmt, args);
        va_end(args);
    }
}

void traceInfo(traceContextData * pContext, char * fmt, ...)
{
    va_list args;

    TRACE_INITIALISE;

    if (g_bSev3Enabled) {
        va_start(args, fmt);
        traceData(pContext, 'I', fmt, args);
        va_end(args);
    }
}

void traceWarning(traceContextData * pContext, char * fmt, ...)
{
    va_list args;

    TRACE_INITIALISE;

    if (g_bSev4Enabled) {
        va_start(args, fmt);
        traceData(pContext, 'W', fmt, args);
        va_end(args);
    }
}

void traceError(traceContextData * pContext, char * fmt, ...)
{
    va_list args;

    TRACE_INITIALISE;

    if (g_bSev5Enabled) {
        va_start(args, fmt);
        traceData(pContext, 'E', fmt, args);
        va_end(args);
    }
}

